/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/* Copyright (C) 2010, Pascal Giorgi
 *
 * Written by Pascal Giorgi pascal.giorgi@lirmm.fr
 *
 * ========LICENCE========
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * ========LICENCE========
 */


#ifndef __CHEBYSHEV_MUL__H
#define __CHEBYSHEV_MUL__H


#include <vector>
#include <algorithm>
#include <complex.h>
#include <fftw3.h>
#include <cassert>

using namespace std;

#define KARA_THRESHOLD 64

// multiplication of Polynomial in Chebyshev Basis C=A*B
// naive quadratic mutliplication
template<class T>
void cheby_mul(vector<T>& C, const vector<T>& A, const vector<T>& B)
{
	size_t dA,dB,dC;
	dA=A.size();
	dB=B.size();
	dC=C.size();	
	assert(dA==dB);
	assert(dA+dB==dC+1);
	
	C[0]=0.5*A[0]*B[0];
	for (size_t i=1;i<dA;i++)
		C[0]+=A[i]*B[i];

	for (size_t i=1;i<dA-1;i++){
		C[i]=0.0;
		for (size_t l=0;l<=i;l++)
			C[i]+=A[i-l]*B[l];
		for (size_t l=1;l<=dA-1-i;l++)
			C[i]+=A[l]*B[l+i]+A[l+i]*B[l];
		C[i]*=0.5;		
	}

	for (size_t i=dA-1;i<dC;i++){
		C[i]=0.0;
		for (size_t l=i-dA+1;l<dA;l++)
			C[i]+=A[i-l]*B[l];
		C[i]*=0.5;
	}		
}

// multiplication of Polynomial in Chebyshev Basis C=A*B
// using DCT-I algorithm to get O(n log n) complexity
template<class T>
void cheby_mul_DCT(vector<T>& C, const vector<T>& A, const vector<T>& B)
{
	size_t dA,dB,dC;
	dA=A.size();
	dB=B.size();
	dC=C.size();	
	assert(dA==dB);
	assert(dA+dB==dC+1);
	
	size_t i, N=dC;
	double *inA, *inB, *inC;
	double *outA, *outB;
	fftw_plan pA,pB,pC;
	
	inA  = new double[N]; outA = new double[N];
	inB  = new double[N]; outB = new double[N];
	inC  = new double[N];
		
	for (i=0;i<dA;i++){ inA[i]=A[i]; inB[i]=B[i];}
		
	for(;i<N;i++){inA[i]=0.0; inB[i]=0.0;}

	pA = fftw_plan_r2r_1d(N, inA,  outA, FFTW_REDFT00, FFTW_ESTIMATE);
	pB = fftw_plan_r2r_1d(N, inB,  outB, FFTW_REDFT00, FFTW_ESTIMATE);
	pC = fftw_plan_r2r_1d(N, inC,  &C[0], FFTW_REDFT00, FFTW_ESTIMATE);
	
	// DCT on inA and inB
	fftw_execute(pA); 
	fftw_execute(pB);

	// pointwise multiplication of double
	for (i=0;i<N;i++)
	  inC[i]= outA[i]*outB[i]/4.0;
	
	// DCT on inC
	fftw_execute(pC);
	
	// scale the result
	double hN=1.0/((double)N-1.0); 
	for (i=0;i<dC;i++)
	  C[i]*=hN;
	C[dC-1]/=2;  
 	
	fftw_destroy_plan(pA);fftw_destroy_plan(pB);fftw_destroy_plan(pC);
	delete[] inA, inB, inC, outA, outB;
}

// multiplication of Polynomial in Chebyshev Basis C=A*B
// template parameter is a function which computes polynomial multiplication in monomial basis
template<class T, void polmul(vector<T>&,const vector<T>&,const vector<T>&)>
void cheby_mul_red(vector<T>& C, const vector<T>& A, const vector<T>& B)
{
	size_t dA,dB,dC;
	dA=A.size();
	dB=B.size();
	dC=C.size();
	assert(dA==dB);
	assert(dA+dB==dC+1);
	vector<T> Bmir(dB),C2(dC);
	
	// take the mirror of polynomial B
	for (size_t i=0;i<dB;i++)
		Bmir[i]=B[dB-1-i];

	// perform the 1st monomial basis multiplication (A*B)
	polmul(C,A,B); 
	
	// scale C by 0.5
	for (size_t i=0;i<dC;++i)
		C[i]*=0.5;
	
	/*
	// perform the 2nd monomial basis multiplication (A*Bmir)
	polmul(C2,A,Bmir);	
	// correct the result C by the value computed in C2
	C[0]+=C2[dA-1]-A[0]*B[0];
	for (size_t i=1;i<dA-1;i++)
		C[i]+= 0.5*(C2[dA-1+i]+C2[dA-1-i]-A[0]*B[i] -A[i]*B[0] );
	*/
	
	/*
	  Set constant term  B0 to zero to avoid subtratcion with A[i]*B[0]
	  could be the same for A, but it would need to modify A wich is const
	*/

	// perform the 2nd monomial basis multiplication (A*Bmir)
	Bmir[dB-1]=0.;
	polmul(C2,A,Bmir);	
	
	// correct the result C by the value computed in C2
	C[0]+=C2[dA-1];//-A[0]*B[0];
	for (size_t i=1;i<dA-1;i++)
		C[i]+= 0.5*(C2[dA-1+i]+C2[dA-1-i]-A[0]*B[i]);	
}




// quadratic monomial basis multiplication
template<class T>
void naive_mul(vector<T>& C, const vector<T>& A, const vector<T>& B)
{
	size_t dA,dB,dC;
	dA=A.size();
	dB=B.size();
	dC=C.size();	

	assert(dA==dB);
	assert(dA+dB==dC+1);
	

	for (size_t i=0;i<dC;i++) C[i]=0.0;
	
	for (size_t i=0;i<dA;i++)
		for (size_t j=0;j<dB;j++)
			C[i+j]+=A[i]*B[j];						
}



// Karatsuba monomial basis multiplication
template <class Iterator, class const_Iterator>
void rec_kara(Iterator c, size_t nc, const_Iterator a, size_t na, const_Iterator b, size_t nb, Iterator buffer)
{
	
	size_t n=na,i;       
	if (n<=KARA_THRESHOLD){
		//*c=(*a)*(*b);	
		for (i=0;i<nc;i++) c[i]=0.0;
		for (i=0;i<na;i++)
			for (size_t j=0;j<nb;j++)
				c[i+j]+=a[i]*b[j];
	}
	else {
		size_t n2 = n>>1;

		//Iterator T0,T1,C0,C1,C01;
		//T0 = buffer; 
		//T1 = buffer += n2;
		//C0 = c;	
		//C01= c+n2;
		//C1 = c+n;

		/* (1) compute C0 = a_lo + a_hi */	
		for (i=0;i<n2;i++)
			*(c+i)=*(a+i)+*(a+i+n2);

		/* (2) compute C01 = b_lo + b_hi */	
		for (i=0;i<n2;i++)
			*(c+n2+i)=*(b+i)+*(b+n2+i);

		/* (3) recursively compute  (a_lo+a_hi)*(b_lo+b_hi) in T0 using C1 as extra space */
		rec_kara(buffer, n-1, c, n2, c+n2, n2, c+n); 
		/* (4) recursively compute a_hi*b_hi in C1 using C0 as extra space */
		rec_kara(c+n, n-1, a+n2, n2, b+n2, n2, c);	

		/* (5) subtract a_hi*b_hi to T0 */
		for (i=0;i<n-1;i++)
			*(buffer+i)-=*(c+n+i);
		
		/* (6,7) add T0 into c at the right place n2..n2+n-1 */
		for (i=0;i<n2;i++)
			*(c+n2+i) =*(buffer+i);
		for (;i<n-1;i++)
			*(c+n2+i)+=*(buffer+i);
	
		/* (8) recursively compute a_lo*b_lo in T0 using as C0 extra space */
		rec_kara(buffer, n-1, a, n2, b, n2, c);
		
		/* (9) add a_lo*b_lo to c at the right place 0..n-1 */
		for (i=0;i<n2;i++)    
			*(c+i) =*(buffer+i);
		for (;i<n-1;i++)
			*(c+i)+=*(buffer+i);
		
		/* (10) subtract a_lo*b_lo to c at the right place n2..n2+n-1 */
		for (i=0;i<n-1;i++)
			*(c+n2+i)-=*(buffer+i);
		
	}

 }

// Karatsuba monomial basis multiplication (2nd version)
template <class Iterator, class const_Iterator>
void rec_kara2(Iterator c, size_t nc, const_Iterator a, size_t na, const_Iterator b, size_t nb)
{	
	size_t i;
	size_t n= na;

	if (n<=KARA_THRESHOLD){		
		for (i=0;i<nc;i++) c[i]=0.0;
		for (i=0;i<na;i++)
			for (size_t j=0;j<nb;j++)
				c[i+j]+=a[i]*b[j];
	}
	else {		
		size_t n2= n>>1;
		//double T1[n2];
		//double T2[n2];
		//double T3[n-1];
		double *T1= new double[n2];
		double *T2= new double[n2];
		double *T3= new double[n-1];
	
		/* compute T1 = a_lo + a_hi */
		for (i=0;i<n2;i++)
			T1[i]=a[i]+a[i+n2];
	
		/* compute T2 = b_lo + b_hi */
		for (i=0;i<n2;i++)
			T2[i]=b[i]+b[i+n2];
	
		/* recursively compute T3 = T1 * T2 */
		rec_kara2(T3, n-1, T1, n2, T2, n2);
		
		delete[] T1;
		delete[] T2;

		/* recursively compute a_hi * b_hi into high part of c */
		/* and subtract from T3 */
		rec_kara2(c+n, n-1, a+n2, n2, b+n2, n2);
		for (i=0;i<n-1;i++)
			T3[i]-=c[n+i];
	
		/* recursively compute a_lo*b_lo into low part of c */
		/* and subtract from T3 */
		rec_kara2(c, n-1, a, n2, b, n2);
		for (i=0;i<n-1;i++)
			T3[i]-=c[i];
	
		c[n-1]=0.0;
		/* finally, add T3 * X^{hsa} to c */
		for (i=0;i<n-1;i++)
			c[i+n2]+=T3[i];	

		delete[] T3;
		
	}
}
template<class T>
void kara_mul(vector<T>& C, const vector<T>& A, const vector<T>& B){
	size_t dA,dB,dC;
	dA=A.size();
	dB=B.size();
	dC=C.size();	
	assert(dA==dB);
	assert(dA+dB==dC+1);
	if (dA<=KARA_THRESHOLD){	
		naive_mul(C,A,B);
	}
	else {
		vector<T> buffer(dA);
		rec_kara(C.begin(), dC, A.begin(), dA, B.begin(), dB, buffer.begin());
	}
}

template<class T>
void kara_mul2(vector<T>& C, const vector<T>& A, const vector<T>& B){
	size_t dA,dB,dC;
	dA=A.size();
	dB=B.size();
	dC=C.size();	
	assert(dA==dB);
	assert(dA+dB==dC+1);
	if (dA<=KARA_THRESHOLD){	
		naive_mul(C,A,B);
	}
	else
		rec_kara2(C.begin(), dC, A.begin(), dA, B.begin(), dB);	
}




#endif
